"""
Document your migration
"""

from happy_migrations import Step

jedi_table = Step(
    forward="""
    CREATE TABLE jedi (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL
    );
    """,
    backward="""
    DROP TABLE jedi;
    """
)

sith_table = Step(
    forward="""
    CREATE TABLE sith (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL
    );
    """,
    backward="""
    DROP TABLE sith;
    """
)

rogue_table = Step(
    forward="""
    CREATE TABLE rogue (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL
    );
    """,
    backward="""
    DROP TABLE rogue;
    """
)

__steps__: tuple = jedi_table, sith_table, rogue_table



"""
Document your migration
"""

from happy_migrations import Step

separatist_table = Step(
    forward="""
    CREATE TABLE separatist (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL
    );
    """,
    backward="""
    DROP TABLE separatist;
    """
)

__steps__: tuple = separatist_table,



from pathlib import Path

import pytest

from happy_migrations.sqlite_backend import (
    SQLiteBackend,
    _mig_name_parser,
    MIGRATION_FILE_TEMPLATE,
    HappyIni
)

ZERO_FNAME = "0000_jedi_sith_rogue_tables"
ONE_FNAME = "0001_separatist_table"

ZERO_MIG_NAME = "jedi_sith_rogue_tables"
ONE_MIG_NAME = "separatist_table"

GET_STATUS = """
    SELECT status
    FROM _happy_status
"""

GET_ZERO_MIG_TABLE_NAMES = """
    SELECT name
    FROM sqlite_master
    WHERE type = 'table'
    AND name IN ('jedi', 'sith', 'rogue');
"""


@pytest.fixture
def db() -> SQLiteBackend:
    migs_dir = Path(__file__).parent.resolve() / "migrations"
    db_path = ":memory:"
    db = SQLiteBackend(HappyIni(db_path=db_path, migs_dir=migs_dir))
    db.happy_init()
    return db


@pytest.fixture
def db_temp(tmp_path) -> SQLiteBackend:
    db = SQLiteBackend(HappyIni(db_path=":memory:", migs_dir=tmp_path))
    db.happy_init()
    return db


def test_mig_incorrect_name():
    assert _mig_name_parser("Mar^&*io") == "mar___io"


def test_mig_parser(tmp_path):
    db = SQLiteBackend(HappyIni(db_path=":memory:", migs_dir=tmp_path))
    db.happy_init()
    db.create_mig("mar_io")
    res = db._parse_mig(tmp_path / "0000_mar_io.py")
    query_body = "\n\n    "
    query = res.steps[0]
    assert query.forward == query_body
    assert query.backward == query_body


def test_is_singleton(db, db_temp):
    assert db is db_temp
    assert db._connection is db_temp._connection


def test_happy_init(db):
    res = db._fetchall("""
        SELECT name
        FROM sqlite_master
        WHERE type='table' AND name='_happy_log' OR name='_happy_status';
    """)
    assert res == [('_happy_status',), ('_happy_log',)]


def test_are_separated():
    mig_dir = Path().resolve().parent / "migrations"
    db_path = ":memory:"
    db = SQLiteBackend(HappyIni(db_path=db_path, migs_dir=mig_dir))
    res = db._fetchall("""
        SELECT name
        FROM sqlite_master
        WHERE type='table' AND name='_happy_log' OR name='_happy_status';
    """)
    assert res == []


def test_get_mig_status_exist(db):
    db._add_mig_to_happy_status(mig_id=0, mig_name="mario")
    res = db._get_mig_status("0000_mario")[:-1]
    assert res == ('0000_mario', 'Pending 🟡')


def test_get_mig_status_not_exist(db):
    db._add_mig_to_happy_status(mig_id=0, mig_name="mario")
    res = db._get_mig_status("0000_luigi")
    assert res is None


def test_create_mig(db_temp):
    db_temp._create_mig_file("mario", 0)
    assert len(tuple(db_temp._migs_dir.glob("0000_mario.py"))) == 1


def test_curren_revision_no_mig(db):
    assert db._get_current_revision_id() == -1


def test_current_revision_one_mig(db_temp):
    db_temp.create_mig("mario")
    assert db_temp._get_current_revision_id() == 0


def test_create_mig_file(db_temp):
    db_temp._create_mig_file(mig_name="mario", mig_id=0)
    with open(db_temp._migs_dir / "0000_mario.py", "r") as file:
        assert file.read() == MIGRATION_FILE_TEMPLATE


def test_add_mig_happy_status(db):
    db._add_mig_to_happy_status(mig_id=0, mig_name="mario")
    query = """
        SELECT fname, status
        FROM _happy_status
        WHERE fname = '0000_mario'
    """
    res = db._fetchone(query)
    assert res == ('0000_mario', 'Pending 🟡')


def test_get_pending_migs_names(db_temp):
    db_temp.create_mig("mario")
    db_temp.create_mig("luigi")
    res = [mig.fname for mig in db_temp._get_pending_migs()]
    assert res == ['0000_mario', '0001_luigi']


def test_exec_all_forward_steps(db):
    mig_path = db._migs_dir / (ZERO_FNAME + ".py")
    db._add_mig_to_happy_status(0, "jedi_sith_rogue_tables")
    mig = db._parse_mig(mig_path)
    db._exec_all_forward_steps(mig)
    query = """
        SELECT name
        FROM sqlite_master
        WHERE type = 'table'
        AND name IN ('jedi', 'sith', 'rogue');
    """
    res = db._fetchall(query)
    assert res == [('jedi',), ('sith',), ('rogue',)]


def test_change_happy_status(db):
    db._add_mig_to_happy_status(mig_id=0, mig_name="mario")
    fname = "0000_mario"
    db._change_happy_status(fname, "A")
    query = """
        SELECT status
        FROM _happy_status
        WHERE fname = :fname
    """
    res = db._fetchall(query, {"fname": fname})
    assert res == [("Applied 🟢",)]
    db._change_happy_status(fname, "P")
    res = db._fetchall(query, {"fname": fname})
    assert res == [("Pending 🟡",)]


def test__get_mig_path(db_temp):
    mig_dir = db_temp._migs_dir
    mig_name = "mario"
    res = db_temp._get_mig_path(mig_name)
    assert res == mig_dir / (mig_name + ".py")


def test_apply_mig_from_name(db):
    db._add_mig_to_happy_status(0, "jedi_sith_rogue_tables")
    mig = db._parse_mig(db._get_mig_path("0000_jedi_sith_rogue_tables"))
    db._apply_mig(mig)
    query = """
        SELECT name
        FROM sqlite_master
        WHERE type = 'table'
        AND name IN ('jedi', 'sith', 'rogue');
    """
    res = db._fetchall(query)
    assert res == [('jedi',), ('sith',), ('rogue',)]


def test_apply_all_migs(db):
    db._add_mig_to_happy_status(0, ZERO_MIG_NAME)
    db._add_mig_to_happy_status(1, ONE_MIG_NAME)
    db.apply_all_migs()
    query = """
        SELECT name
        FROM sqlite_master
        WHERE type = 'table'
        AND name IN ('jedi', 'sith', 'rogue', 'separatist');
    """
    res = db._fetchall(query)
    assert res == [('jedi',), ('sith',), ('rogue',), ('separatist',)]


# TODO: Refactor SQL

def test_get_all_migs_up_to_pending(db):
    db._add_mig_to_happy_status(0, ZERO_MIG_NAME)
    db._add_mig_to_happy_status(1, ONE_MIG_NAME)
    res = db._get_all_migs_names_up_to(mig_id=0, status="P", order="ASC")
    assert res[0].fname == '0000_jedi_sith_rogue_tables'


def test_get_all_migs_up_to_applied(db):
    for id, mig_name in enumerate((ZERO_MIG_NAME, ONE_MIG_NAME)):
        db._add_mig_to_happy_status(id, mig_name)
        db._change_happy_status(f"{id:04}_{mig_name}", "A")

    migs = db._get_all_migs_names_up_to(mig_id=0, status="A", order="DESC")
    res = [mig.fname for mig in migs]
    assert res == ['0001_separatist_table', '0000_jedi_sith_rogue_tables']


@pytest.mark.parametrize(
    ("up_to", "result"),
    (
        (0, [('Applied 🟢',), ('Pending 🟡',)]),
        (1, [('Applied 🟢',), ('Applied 🟢',)])
    )
)
def test_apply_migs_up_to_0(db, up_to, result):
    db._add_mig_to_happy_status(0, ZERO_MIG_NAME)
    db._add_mig_to_happy_status(1, ONE_MIG_NAME)
    db.apply_migs_up_to(up_to)
    query = """
        SELECT status
        FROM _happy_status
    """
    res = db._fetchall(query)
    assert res == result


@pytest.mark.parametrize(
    ("up_to", "result"),
    (
            (0, [('Pending 🟡',), ('Pending 🟡',)]),
            (1, [('Applied 🟢',), ('Pending 🟡',)])
    )
)
def test_rollback_migs_up_to(db, up_to, result):
    db._add_mig_to_happy_status(0, ZERO_MIG_NAME)
    db._add_mig_to_happy_status(1, ONE_MIG_NAME)
    db.apply_all_migs()
    db.rollback_migs_up_to(up_to)
    query = """
        SELECT status
        FROM _happy_status
    """
    res = db._fetchall(query)
    assert res == result


def test_rollback_mig_true(db):
    db._add_mig_to_happy_status(0, ZERO_MIG_NAME)
    db.apply_all_migs()
    assert db._fetchall(GET_STATUS) == [('Applied 🟢',)]
    assert db.rollback_last_mig()
    res = db._fetchall(GET_STATUS)
    assert res == [('Pending 🟡',)]


def test_rollback_mig_false(db):
    assert not db.rollback_last_mig()


def test_exec_all_backward_steps(db):
    mig_path = db._migs_dir / (ZERO_FNAME + ".py")
    db._add_mig_to_happy_status(0, "jedi_sith_rogue_tables")
    mig = db._parse_mig(mig_path)
    db._exec_all_forward_steps(mig)
    db._exec_all_backward_steps(mig)
    res = db._fetchall(GET_ZERO_MIG_TABLE_NAMES)
    assert res == []



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path


@dataclass
class HappyIni:
    db_path: Path | str
    migs_dir: Path

    def __post_init__(self):
        if isinstance(self.db_path, str) and self.db_path != ":memory:":
            self.db_path = Path(self.db_path)
        if isinstance(self.migs_dir, str):
            self.migs_dir = Path(self.migs_dir)


@dataclass
class Step:
    forward: str
    backward: str


@dataclass
class Migration:
    fname: str
    status: str
    created: datetime
    steps: tuple[Step, ...]


@dataclass
class HappyLog:
    id_happy_log: int
    mig_id: int
    operation: str
    username: str
    hostname: str
    created: datetime



import importlib.util
import re
from pathlib import Path
from sqlite3 import connect, Connection, Cursor
from typing import Literal

from happy_migrations import Migration
from happy_migrations._data_classes import HappyIni


MIGRATION_FILE_TEMPLATE = """\
\"\"\"
Document your migration
\"\"\"

from happy_migrations import Step

first_step = Step(
    forward=\"\"\"

    \"\"\",
    backward=\"\"\"

    \"\"\"
)

__steps__: tuple = first_step,
"""

MIG_IS_NOT_TUPLE = "__steps__ is not a tuple inside migration: "


CREATE_HAPPY_STATUS_TABLE = """
CREATE TABLE IF NOT EXISTS _happy_status (
    id_happy_status integer primary key autoincrement,
    mig_id integer,
    fname varchar(255),
    status varchar(255),
    created TIMESTAMP NOT NULL DEFAULT current_timestamp
);
"""

ADD_HAPPY_STATUS = """
INSERT INTO _happy_status (mig_id, fname, status)
VALUES (:mig_id, :fname, :status)
"""

CREATE_HAPPY_LOG_TABLE = """
CREATE TABLE IF NOT EXISTS _happy_log (
    id_happy_log integer primary key autoincrement,
    mig_id integer,
    operation varchar(255),
    username varchar(255),
    hostname varchar(255),
    created TIMESTAMP NOT NULL DEFAULT current_timestamp
);
"""

ADD_HAPPY_LOG = """
INSERT INTO _happy_log (mig_id, operation, username, hostname)
VALUES (:mig_id, :operation, :username, :hostname)
"""

GET_CURRENT_REVISION = """
SELECT fname
FROM _happy_status
ORDER BY fname DESC
LIMIT 1
"""

GET_PENDING_MIGS_NAMES = """
SELECT fname
FROM _happy_status
WHERE status = :status
ORDER BY fname
"""

GET_MIGS_UP_TO = """
SELECT fname
FROM _happy_status
WHERE
    (mig_id <= CASE WHEN :order = 'ASC' THEN :mig_id END
    OR
    mig_id >= CASE WHEN :order = 'DESC' THEN :mig_id END)
AND status = :status
ORDER BY
    CASE WHEN :order = 'ASC' THEN mig_id END,
    CASE WHEN :order = 'DESC' THEN mig_id END DESC;
"""

UPDATE_HAPPY_STATUS = """
UPDATE _happy_status
SET status = :status
WHERE fname = :fname
"""

GET_LAST_BY_STATUS = """
SELECT fname
FROM _happy_status
WHERE status = ?
ORDER BY mig_id DESC
LIMIT 1
"""

LIST_HAPPY_STATUS = """
SELECT fname, status, created
FROM _happy_status
"""

GET_HAPPY_STATUS_BY_FNAME = """
SELECT fname, status, created
FROM _happy_status
WHERE fname = :fname
"""

LIST_HAPPY_LOG = """
SELECT mig_id, operation, username, hostname, created
FROM _happy_log
"""

HAPPY_STATUS = {
    "A": "Applied 🟢",
    "P": "Pending 🟡",
}

INI_TEMPLATE = """\
[HAPPY]
db_path = path\\to\\db
"""


class MigrationError(Exception):
    pass


def _mig_name_parser(string: str) -> str:
    """Converts a given string to a normalized migration name format."""
    return re.sub(r'[^a-zA-Z0-9_]', '_', string).lower()


class SQLiteBackend:
    _instance = None

    def __new__(cls, *args, **kwargs) -> "SQLiteBackend":
        """Creates a new instance if one does not already exist; otherwise,
        returns the existing instance.
        Ensures the class follows the Singleton pattern.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, happy: HappyIni) -> None:
        self._config = happy
        self._migs_dir = happy.migs_dir
        self._db_path = happy.db_path
        self._connection: Connection = connect(self._db_path)

    def _execute(self, query: str, params: dict | tuple = ()) -> Cursor:
        """Execute a SQL query with optional parameters and return a cursor."""
        return self._connection.execute(query, params)

    def _fetchone(self, query: str, params: dict | tuple = ()) -> tuple | None:
        """Execute a SQL query and fetches the first row of the result."""
        return self._execute(query=query, params=params).fetchone()

    def _fetchall(self, query: str, params: dict | tuple = ()) -> list:
        """Execute a SQL query and fetches all rows of the result."""
        return self._execute(query=query, params=params).fetchall()

    def _commit(self):
        """Commit the current transaction to the database."""
        self._connection.commit()

    def _reconnect(self):
        """Reconnect connection to DB."""
        self._connection.close()
        self._connection = connect(self._db_path)

    def _get_mig_status(self, fname: str) -> list | None:
        """Return mig status row if mig exist."""
        params = {"fname": fname}
        return self._fetchone(GET_HAPPY_STATUS_BY_FNAME, params)

    def _parse_mig(self, mig_path: Path) -> Migration:
        """Parses a migration file and returns a `Migration` object."""
        spec = importlib.util.spec_from_file_location(
            mig_path.name, mig_path
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        queries = getattr(module, "__steps__")
        if not isinstance(queries, tuple):
            raise ValueError(MIG_IS_NOT_TUPLE + mig_path.name)
        return Migration(*self._get_mig_status(mig_path.stem), steps=queries)

    def happy_init(self) -> None:
        """Initializes the Happy migration system by verifying the migrations
        dir exist and create necessary tables if needed.
        """
        self._migs_dir.mkdir(parents=True, exist_ok=True)
        self._execute(CREATE_HAPPY_STATUS_TABLE)
        self._execute(CREATE_HAPPY_LOG_TABLE)
        self._commit()

    def happy_boot(self) -> None:
        """Initializes Happy and applies all migrations.
        Required during app startup to integrate Happy into the app.
        """
        # TODO: Write test
        self.happy_init()
        self.apply_all_migs()
        self.close_connection()

    def _get_current_revision_id(self) -> int:
        """Retrieves the latest migration revision id from the database
        or return -1 if empty.
        """
        fname: list[str] | None = self._fetchone(GET_CURRENT_REVISION)
        if not fname:
            return -1
        return int(fname[0].split("_", maxsplit=1)[0])

    def create_mig(self, mig_name: str) -> None:
        """Create new migration."""
        mig_name = _mig_name_parser(mig_name)
        mig_id: int = self._get_current_revision_id() + 1
        self._create_mig_file(mig_name=mig_name, mig_id=mig_id)
        self._add_mig_to_happy_status(mig_id=mig_id, mig_name=mig_name)

    def _create_mig_file(self, mig_name: str, mig_id: int) -> None:
        """Create new boilerplate migration file."""
        name = f"{mig_id:04}_{mig_name}.py"
        with open(self._migs_dir / name, 'w') as file:
            file.write(MIGRATION_FILE_TEMPLATE)

    def _add_mig_to_happy_status(self, mig_id: int, mig_name: str) -> None:
        """Add new migration to db with status pending."""
        params = {
            "mig_id": mig_id,
            "fname": f"{mig_id:04}_{mig_name}",
            "status": HAPPY_STATUS['P'],
        }
        self._connection.execute(ADD_HAPPY_STATUS, params)
        self._commit()

    def _get_pending_migs(self) -> list[Migration]:
        """Return all pending migrations' names."""
        params = {"status": HAPPY_STATUS["P"]}
        rows = self._fetchall(GET_PENDING_MIGS_NAMES, params)
        return [self._parse_mig(self._get_mig_path(row[0])) for row in rows]

    def _exec_all_forward_steps(self, mig: Migration) -> None:
        """Execute every forward Query from a Migration."""
        for query in mig.steps:
            self._execute(query.forward)

    def _get_mig_path(self, fname: str) -> Path:
        """Return full path to migration."""
        return self._migs_dir / (fname + ".py")

    def _change_happy_status(self, fname: str, status: Literal["A", "P"]) -> None:
        """Updates the `applied` status of a specific migration
        in the `_happy_status` database table.
        """
        params = {"status": HAPPY_STATUS[status], "fname": fname}
        self._execute(UPDATE_HAPPY_STATUS, params)

    def _apply_mig(self, mig: Migration) -> None:
        self._exec_all_forward_steps(mig)
        self._change_happy_status(mig.fname, "A")
        self._commit()

    def _rollback_mig(self, mig: Migration) -> None:
        self._exec_all_backward_steps(mig)
        self._change_happy_status(mig.fname, "P")
        self._commit()

    def _get_all_migs_names_up_to(
        self,
        mig_id: int,
        status: Literal["A", "P"],
        order: Literal["ASC", "DESC"]
    ) -> list[Migration]:
        """Retrieves all pending migration names from
        the `_happy_status` table up to a specified migration ID.
        """
        rows = self._fetchall(
            GET_MIGS_UP_TO,
            {"mig_id": mig_id, "status": HAPPY_STATUS[status], "order": order}

        )
        return [self._parse_mig(self._get_mig_path(row[0])) for row in rows]

    def _exec_all_backward_steps(self, mig: Migration) -> None:
        """Rolls back a migration by executing each backward SQL statement
        from the last Query to the first.
        """
        for query in mig.steps[::-1]:
            self._execute(query.backward)

    def close_connection(self):
        """Close connection to DB."""
        self._connection.close()

    def apply_last(self):
        """
        Applies the last migration if available.
        If no migration is found, prints a message to the console.
        """
        pass

    def rollback_last(self):
        """
        Rolls back the last applied migration if available.
        If no migration is found, prints a message to the console.
        """
        pass

    def apply_all_migs(self) -> None:
        """Apply all migrations."""
        migs = self._get_pending_migs()
        for mig in migs:
            self._apply_mig(mig)

    def apply_migs_up_to(self, mig_id: int) -> None:
        """Applies all pending migrations up to the specified migration ID."""
        migs = self._get_all_migs_names_up_to(mig_id, "P", "ASC")
        for mig in migs:
            self._apply_mig(mig)

    def rollback_migs_up_to(self, mig_id: int) -> None:
        """Roll back all applied migrations up to the specified migration ID."""
        migs = self._get_all_migs_names_up_to(mig_id, "A", "DESC")
        for mig in migs:
            self._rollback_mig(mig)

    def rollback_last_mig(self) -> bool:
        """Roll back the last applied migration and return True.
        If no migration is available to roll back return False.
        """
        name = self._fetchone(
            GET_LAST_BY_STATUS,
            (HAPPY_STATUS["A"],)
        )
        if name is None:
            return False
        mig_path = self._get_mig_path(name[0])
        mig = self._parse_mig(mig_path)
        self._exec_all_backward_steps(mig)
        self._change_happy_status(mig.fname, "P")
        self._commit()
        return True

    def list_happy_logs(self) -> list:
        """Return list of all logs from _happy_log."""
        # TODO: Test when table is ok
        return self._fetchall(LIST_HAPPY_LOG)

    def list_happy_status(self) -> list:
        """Return list of all statuses from _happy_status."""
        # TODO: Test when table is ok
        return self._fetchall(LIST_HAPPY_STATUS)



import configparser
from pathlib import Path
from typing import cast

from happy_migrations._data_classes import HappyIni


_HAPPY_INI_TEMPLATE = """\
[Settings]
db_path =
migs_dir =
theme = dark
"""


def create_happy_ini(path: Path) -> bool:
    """Create happy.ini file in CWD."""
    if path.exists():
        return True

    with open(path, "w") as file:
        file.write(_HAPPY_INI_TEMPLATE)

    return False


def parse_happy_ini() -> HappyIni:
    """Parse the 'happy.ini' configuration file and return a HappyIni dataclass instance."""
    config = configparser.ConfigParser()
    config.read('happy.ini')
    return HappyIni(
        db_path=cast(Path, config['Settings']['db_path']),
        migs_dir=cast(Path, config['Settings']['migs_dir'])
    )



from pathlib import Path

import click
from click import echo, style

from happy_migrations import (
    SQLiteBackend,
    parse_happy_ini,
    create_happy_ini
)
from happy_migrations._textual_app import StatusApp


@click.group()
def happy() -> None:
    """Happy CLI."""
    pass


@click.command()
def config():
    """Create happy.ini file inside CWD."""
    message = "Happy.ini already exist."
    path = Path().cwd() / "happy.ini"

    if create_happy_ini(path):
        echo(style("Warning: ", "yellow") + message)
    else:
        echo(style("Created: ", "green") + str(path))


@click.command()
def init() -> None:
    """Initializes the Happy migrations."""
    db = SQLiteBackend(parse_happy_ini())
    db.happy_init()
    db.close_connection()


@click.command()
@click.argument("migration_name")
def cmig(migration_name: str) -> None:
    """Create migration."""
    db = SQLiteBackend(parse_happy_ini())
    db.create_mig(mig_name=migration_name)
    db.close_connection()


@click.command()
def log() -> None:
    """Display _happy_log table."""
    pass


@click.command()
def status() -> None:
    """Display _happy_status table."""
    db = SQLiteBackend(parse_happy_ini())
    StatusApp(
        headers=["Name", "Status", "Creation Date"],
        rows=db.list_happy_status()
    ).run(inline=True, inline_no_clear=True)
    db.close_connection()


@click.command()
def fixture():
    """Create 1000 migrations with names based on 孫子 quotes names."""
    from random import randint
    quotes = [
        "all_warfare_is_based_on_deception",
        "the_wise_warrior_avoids_the_battle",
        "in_the_midst_of_chaos_opportunity",
        "move_swift_as_the_wind",
        "strategy_without_tactics_is_slow",
        "let_your_plans_be_dark",
        "supreme_art_is_to_subdue",
        "opportunities_multiply_as_they_are_seized",
        "he_will_win_who_knows_when_to_fight",
        "quickness_is_the_essence_of_war"
    ]
    db = SQLiteBackend(parse_happy_ini())
    for _ in range(10**3):
        db.create_mig(quotes[randint(0, 9)])
    db.close_connection()


happy.add_command(init)
happy.add_command(cmig)
happy.add_command(config)
happy.add_command(log)
happy.add_command(status)
happy.add_command(fixture)



"""happy_migrations' __init__.py"""

from ._data_classes import Step, Migration
from .sqlite_backend import SQLiteBackend
from .utils import parse_happy_ini, create_happy_ini

__all__ = [
    "Migration",
    "Step",
    "SQLiteBackend",
    "parse_happy_ini",
    "create_happy_ini",
]



from textual import events
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import DataTable, Static


class StatusApp(App):
    BINDINGS = (
        Binding("j", "cursor_up_5", "Cursor up", show=False),
        Binding("k", "cursor_down_5", "Cursor down", show=False),
        Binding("h", "cursor_up_20", "Cursor up", show=False),
        Binding("l", "cursor_down_20", "Cursor down", show=False),
    )
    TABLE = DataTable()

    def action_cursor_up_20(self):
        for _ in range(4):
            self.action_cursor_up_5()

    def action_cursor_down_20(self):
        for _ in range(4):
            self.action_cursor_down_5()

    def action_cursor_up_5(self):
        for _ in range(5):
            self.TABLE.action_cursor_up()

    def action_cursor_down_5(self):
        for _ in range(5):
            self.TABLE.action_cursor_down()

    def __init__(self, headers, rows, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = headers
        self.rows = rows

    def compose(self) -> ComposeResult:
        self.TABLE.add_columns(*self.headers)
        self.TABLE.add_rows(self.rows)
        self.TABLE.zebra_stripes = True
        self.TABLE.cursor_type = "row"
        with Static():
            yield self.TABLE

    async def _on_key(self, event: events.Key) -> None:
        if event.name == "escape":
            self.clear_notifications()
            self.exit()
